/*BEGIN_COPYRIGHT_BLOCK
 *
 * This file is part of DrJava.  Download the current version of this project from http://www.drjava.org/
 * or http://sourceforge.net/projects/drjava/
 *
 * DrJava Open Source License
 * 
 * Copyright (C) 2001-2006 JavaPLT group at Rice University (javaplt@rice.edu).  All rights reserved.
 *
 * Developed by:   Java Programming Languages Team, Rice University, http://www.cs.rice.edu/~javaplt/
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
 * documentation files (the "Software"), to deal with the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and 
 * to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 * 
 *     - Redistributions of source code must retain the above copyright notice, this list of conditions and the 
 *       following disclaimers.
 *     - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the 
 *       following disclaimers in the documentation and/or other materials provided with the distribution.
 *     - Neither the names of DrJava, the JavaPLT, Rice University, nor the names of its contributors may be used to 
 *       endorse or promote products derived from this Software without specific prior written permission.
 *     - Products derived from this software may not be called "DrJava" nor use the term "DrJava" as part of their 
 *       names without prior written permission from the JavaPLT group.  For permission, write to javaplt@rice.edu.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
 * THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF 
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
 * WITH THE SOFTWARE.
 * 
 *END_COPYRIGHT_BLOCK*/

package edu.rice.cs.drjava.model.compiler;

import java.io.IOException;
import java.io.File;
import java.util.List;
import edu.rice.cs.drjava.model.OpenDefinitionsDocument;

/** Interface for all compiler functionality in the model.  The compilation process itself can be monitored through
 *  the CompilerListener interface.  The four primary uses of this interface will be to manage listeners, to trigger
 *  compilation of (a) document(s), to handle the results, and to manage available compilers.
 * 
 *  @version $Id: CompilerModel.java 4195 2007-03-01 20:12:36Z dlsmith $
 */
public interface CompilerModel {
  
  /** Legal source file name extensions */
  public static final String[] EXTENSIONS = new String[]{".java", ".dj0", ".dj1", ".dj2"};
  
  //----------------------------Locking--------------------------------------//
  
  /** Returns the lock used to prevent simultaneous compilation and JUnit testing */
  public Object getCompilerLock();
  
  //-------------------------- Listener Management --------------------------//
  
  /** Add a CompilerListener to the model.
   *  @param listener a listener that reacts to compiler events
   */
  public void addListener(CompilerListener listener);

  /** Remove a CompilerListener from the model.  If the listener is not currently listening to this model, this method
   *  has no effect.
   *  @param listener a listener that reacts to compiler events
   */
  public void removeListener(CompilerListener listener);

  /** Removes all CompilerListeners from this model. */
  public void removeAllListeners();
  
  //-------------------------------- Triggers --------------------------------//
  
  /** Compiles all documents, which requires that the documents be saved first.
   *  @throws IOException if a filesystem-related problem prevents compilation
   */
  public void compileAll() throws IOException;
  
  /** Compiles all documents in the project source tree, which requires that the documents be saved first.
   *  @throws IOException if a filesystem-related problem prevents compilation
   */
  public void compileProject() throws IOException;
  
  /** Compiles the specified documents which must be saved first.
   *  @param docs the documents to be compiled
   *  @throws IOException if a filesystem-related problem prevents compilation
   */
  public void compile(List<OpenDefinitionsDocument> docs) throws IOException;
  
  /** Compiles a single document which must be saved first.
   *  @param doc the document to be compiled
   *  @throws IOException if a filesystem-related problem prevents compilation
   */
  public void compile(OpenDefinitionsDocument doc) throws IOException;
  
  //----------------------------- Error Results -----------------------------//
  
  /** Gets the CompilerErrorModel representing the last compile. */
  public CompilerErrorModel getCompilerErrorModel();
  
  /** Gets the total number of current errors. */
  public int getNumErrors();
  
  /** Resets the compiler error state to have no errors. */
  public void resetCompilerErrors();
  
  //-------------------------- Compiler Management --------------------------//

  /**
   * Returns all registered compilers that are actually available.  If there are none,
   * the result is {@link NoCompilerAvailable#ONLY}.
   */
  public Iterable<CompilerInterface> getAvailableCompilers();

  /**
   * Gets the compiler that is the "active" compiler.
   *
   * @see #setActiveCompiler
   */
  public CompilerInterface getActiveCompiler(); 

  /**
   * Sets which compiler is the "active" compiler.
   *
   * @param compiler Compiler to set active.
   * @throws IllegalArgumentException  If the compiler is not in the list of available compilers
   *
   * @see #getActiveCompiler
   */
  public void setActiveCompiler(CompilerInterface compiler);
  
  /** Add a compiler to the active list */
  public void addCompiler(CompilerInterface compiler);
  
}
